__VERSION_MAJOR__ = '1'
__VERSION_MINOR__ = '0'
__VERSION_BUILD__ = '1'
__VERSION_TAGS__  = ''

__VERSION_LIST__ = [
    __VERSION_MAJOR__,
    __VERSION_MINOR__,
    __VERSION_BUILD__,
    __VERSION_TAGS__]

__VERSION__ = f'{__VERSION_MAJOR__}.{__VERSION_MINOR__}.{__VERSION_BUILD__}{__VERSION_TAGS__}'

__DATA_VERSION__ = 1

from audata.file import File
